"""Класс управления снимками ВМ."""

import os
import time
import signal
import subprocess
import uuid
import shlex
import xml.etree.ElementTree as ET
import logging

from msbackup.exception import LibvirtError


logger = logging.getLogger("msbackup")


def preexec_fn():  # pragma: no coverage
    """Отправка сигнала подпроцессу."""
    signal.signal(signal.SIGPIPE, signal.SIG_DFL)


class Snapshot:
    """Снимок виртуальной машины."""

    def __init__(self, conn, domain_name, pool=None):
        """Конструктор."""
        self.domain_name = domain_name
        self.snapshot_xml = None
        self.disks = None
        self.snapshot_disk = None
        self.snapshot_id = None
        self.conn = conn
        self.snapshot = None
        pool_name = pool if pool is not None else "default"
        self.pool = self.conn.storagePoolLookupByName(pool_name)
        root = ET.fromstring(self.pool.XMLDesc())
        self.pool_path = root.find("./target/path").text

    def get_domain(self):
        """Получение виртуальной машины по имени домена."""
        return self.conn.lookupByName(self.domain_name)

    def is_active(self):
        """Проверка активности виртуальной машины."""
        return self.get_domain().isActive() != 0

    def get_disks(self):
        """Получение всех дисков виртуальной машины."""
        domain = self.get_domain()
        root = ET.fromstring(domain.XMLDesc())
        devices = root.findall("./devices/disk[@device='disk']")
        sources = [device.find("source").attrib for device in devices]
        targets = [device.find("target").attrib for device in devices]
        drivers = [device.find("driver").attrib for device in devices]
        if len(sources) != len(targets):  # pragma: no coverage
            raise RuntimeError(
                "Targets and sources lengths are different %s:%s"
                % (len(sources), len(targets))
            )
        disks = {}
        for i in range(len(sources)):
            if drivers[i]["type"] == "qcow2":
                disks[targets[i]["dev"]] = {
                    "file": (
                        sources[i]["file"]
                        if "file" in sources[i]
                        else sources[i]["dev"]
                    ),
                    "driver": drivers[i]["type"],
                    "snapshot": True,
                }
            else:
                disks[targets[i]["dev"]] = {
                    "file": "",
                    "driver": "",
                    "snapshot": False,
                }
        return disks

    def dump_xml(self, path):
        """Сохранение настроек виртуальной машины."""
        import libvirt

        domain = self.get_domain()
        logger.debug("Dumping XMLs for domain %s", domain.name())
        xml_files = []
        dest_file = os.path.join(path, "%s.xml" % domain.name())
        ERROR_MSG = "File %s already exists!"
        if os.path.exists(dest_file):  # pragma: no coverage
            raise RuntimeError(ERROR_MSG % dest_file)
        with open(dest_file, mode="w") as dest_fh:
            # dump different xmls files. First of all, the offline dump
            xml = domain.XMLDesc()
            dest_fh.write(xml)
        xml_files += [dest_file]
        DEBUG_MSG = "File %s wrote"
        logger.debug(DEBUG_MSG, dest_file)
        # All flags:
        # - libvirt.VIR_DOMAIN_XML_INACTIVE
        # - libvirt.VIR_DOMAIN_XML_MIGRATABLE
        # - libvirt.VIR_DOMAIN_XML_SECURE
        # - libvirt.VIR_DOMAIN_XML_UPDATE_CPU
        dest_file = "%s-inactive.xml" % domain.name()
        dest_file = os.path.join(path, dest_file)
        if os.path.exists(dest_file):  # pragma: no coverage
            raise RuntimeError(ERROR_MSG % dest_file)
        with open(dest_file, mode="w") as dest_fh:
            # dump different xmls files. First of all, the offline dump.
            xml = domain.XMLDesc(flags=libvirt.VIR_DOMAIN_XML_INACTIVE)
            dest_fh.write(xml)
        xml_files += [dest_file]
        logger.debug(DEBUG_MSG, dest_file)
        # Dump a migrate config file
        dest_file = "%s-migratable.xml" % domain.name()
        dest_file = os.path.join(path, dest_file)
        if os.path.exists(dest_file):  # pragma: no coverage
            raise RuntimeError(ERROR_MSG % dest_file)
        with open(dest_file, mode="w") as dest_fh:
            # dump different xmls files. First of all, the offline dump.
            xml = domain.XMLDesc(
                flags=(
                    libvirt.VIR_DOMAIN_XML_INACTIVE + libvirt.VIR_DOMAIN_XML_MIGRATABLE
                )
            )
            dest_fh.write(xml)
        xml_files += [dest_file]
        logger.debug(DEBUG_MSG, dest_file)
        return xml_files

    def has_current_snapshot(self):
        """Проверка наличия снимка виртуальной машины."""
        return self.get_domain().hasCurrentSnapshot() != 0

    def get_xml(self):
        """Получение состояния дисков виртуальной машины в формате XML."""
        domain = self.get_domain()
        self.disks = self.get_disks()
        self.snapshot_id = str(uuid.uuid1()).split("-")[0]
        diskspecs = []
        fmt = "--diskspec %s,file=%s/snapshot_%s_%s-%s.img"
        for disk, spec in self.disks.items():
            if spec["snapshot"]:
                diskspecs.append(
                    fmt
                    % (
                        disk,
                        self.pool_path,
                        self.domain_name,
                        disk,
                        self.snapshot_id,
                    )
                )
            else:
                diskspecs.append(f"--diskspec {disk},snapshot=no")
        ds = " ".join(diskspecs)
        my_cmd = (
            f"virsh snapshot-create-as --domain {domain.name()} "
            f"{self.snapshot_id} {ds} --disk-only --atomic --quiesce --print-xml"
        )
        logger.debug("Executing: %s", my_cmd)
        create_xml = subprocess.Popen(
            shlex.split(my_cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=preexec_fn,
            shell=False,
        )
        self.snapshot_xml = create_xml.stdout.read().decode("utf-8")
        status = create_xml.wait()
        if status != 0:  # pragma: no coverage
            logger.error("Error for %s:%s", my_cmd, create_xml.stderr.read())
            logger.critical("%s returned %s state", "virsh", status)
            raise RuntimeError("snapshot-create-as didn't work properly")
        return self.snapshot_xml

    def create_snapshot(self):
        """Создание снимка состояния виртуальной машины."""
        import libvirt

        domain = self.get_domain()
        if self.snapshot_xml is None:
            self.get_xml()
        if self.is_active():
            if self.snapshot is not None:
                logger.error("A snapshot is already defined for this domain")
                logger.warning("Returning the current snapshot")
                return self.snapshot
            logger.debug(
                f"Creating snapshot {self.snapshot_id} for domain {self.domain_name}"
            )
            self.snapshot = domain.snapshotCreateXML(
                self.snapshot_xml,
                flags=sum(
                    [
                        libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_DISK_ONLY,
                        libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_ATOMIC,
                        libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_QUIESCE,
                    ]
                ),
            )
        else:
            logger.debug("Domain inactive - skip create snapshot")
        self.snapshot_disk = self.get_disks()
        for disk, info in self.snapshot_disk.items():
            if info["snapshot"]:
                top = info["file"]
                logger.debug(
                    f"Created top image {top} for disk {disk} of domain {domain.name()}"
                )
        return self.snapshot

    def do_blockcommit(self):
        """Выполнение команды blockcommit для каждого диска в снимке."""
        import libvirt

        if not self.is_active():
            logger.debug("Domain inactive - skip blockcommit")
            return False
        if self.snapshot is None:
            raise LibvirtError(f"no snapshot of domain: {self.domain_name}")
        domain = self.get_domain()
        logger.debug("Block-committing domain: %s", domain.name())
        bc_flags = libvirt.VIR_DOMAIN_BLOCK_COMMIT_ACTIVE
        piv_flags = libvirt.VIR_DOMAIN_BLOCK_JOB_ABORT_PIVOT
        for disk, spec in self.disks.items():
            if not spec["snapshot"]:
                continue
            info = self.snapshot_disk[disk]
            logger.debug("block commit image %s in disk %s", info["file"], disk)
            try:
                domain.blockCommit(disk, None, info["file"], 0, bc_flags)
                logger.debug("snapshot %s merged", self.snapshot_id)
            except libvirt.libvirtError as e:
                logger.error("failed to merge snapshot: %s", e)
                return False
            # Wait for job complete
            while True:
                time.sleep(1)
                job = domain.blockJobInfo(disk, 0)
                logger.debug("block job info: %s", job)
                if job["cur"] == job["end"]:
                    break
            try:
                domain.blockJobAbort(disk, piv_flags)
            except libvirt.libvirtError as e:
                logger.error("failed to pivot snapshot: %s", self.snapshot_id)
                return False
            return True

    def remove_snapshot(self):
        """Удаление снимков дисков и метаданных снимка ВМ."""
        if not self.is_active():
            logger.debug("Domain inactive - skip blockcommit")
            return

        import libvirt

        test_disks = self.get_disks()
        for disk, spec in self.disks.items():
            if not spec["snapshot"]:
                continue
            base = spec["file"]
            test_base = test_disks[disk]["file"]
            top = self.snapshot_disk[disk]["file"]
            if base == test_base and top != test_base:
                logger.debug("Removing %s", top)
                self.pool.refresh()
                vol = self.pool.storageVolLookupByName(os.path.basename(top))
                vol.delete()
            else:  # pragma: no coverage
                logger.error(
                    "original base: %s, top: %s, new_base: %s",
                    base,
                    top,
                    test_base,
                )
                raise RuntimeError(
                    "Something goes wrong for snaphost",
                    self.snapshot_id,
                )
        logger.debug("Removing snapshot %s", self.snapshot_id)
        metadata = [libvirt.VIR_DOMAIN_SNAPSHOT_DELETE_METADATA_ONLY]
        self.snapshot.delete(flags=sum(metadata))
