"""Модуль архиватора баз данных SQLite."""

import os
import subprocess
import shlex
import logging

from msbackup.backend_base import Base as BaseBackend
from msbackup.exception import ArchiveError


logger = logging.getLogger("msbackup")


def get_backend_kwargs(params):
    """Подготовка параметров для режима 'sqlite'."""
    kwargs = BaseBackend.get_common_backend_kwargs(params)
    BaseBackend.get_param(params, kwargs, "sqlite_cmd")
    return kwargs


class Sqlite(BaseBackend):
    """Архиватор баз данных SQLite."""

    SECTION = "Backend-SQLite"

    @classmethod
    def make_subparser(cls, subparsers):
        """Добавление раздела параметров командной строки для архиватора."""
        parser = subparsers.add_parser("sqlite")
        parser.set_defaults(get_backend_kwargs=get_backend_kwargs)
        parser.add_argument(
            "--sqlite-cmd",
            help="Command to run SQLite command-line shell.",
        )
        parser.add_argument(
            "source", nargs="*", metavar="DBNAME", help="Path to database file."
        )

    def __init__(self, config, *args, **kwargs):
        """Конструктор."""
        super().__init__(config, *args, **kwargs)
        self.suffix = ".gz" + (self.encryptor.suffix if self.encryptor else "")
        self.sqlite_cmd = shlex.split(
            kwargs.get("sqlite_cmd")
            or config.get(
                section=self.SECTION,
                option="SQLITE_COMMAND",
                fallback="/usr/bin/sqlite3",
            )
        )
        self.tmp_db_list = []
        self.tmp_file_list = []

    def source_name(self, source):
        """Имя источника архива."""
        return os.path.splitext(source)[0]

    def _backup(self, sources=None, **kwargs):
        """
        Архивация баз данных SQLite.

        :param source: Список источников для архивации.
        :type source: [str]
        :return: Количество ошибок.
        :rtype: int
        """
        if sources is None:
            raise AttributeError("source is not specified")
        error_count = 0
        for source in sources:
            output = self.outpath(os.path.basename(source))
            logger.info('Backup source "%s".', source)
            try:
                self.archive(
                    source=source,
                    output=output,
                )
            except subprocess.CalledProcessError as ex:
                logger.error(str(ex))
                error_count += 1
        return error_count

    def _archive(self, source, output=None, **kwargs):
        """
        Упаковка списка источников в файл архива.

        :param source: Путь к файлу БД SQLite.
        :type source: str
        :param output: Путь к файлу архива.
        :type output: str
        :return: Процесс упаковки файлов.
        :rtype: subprocess.Popen
        """
        tmp_db = os.path.join(self.tmp_dir, os.path.basename(source))
        self.tmp_db_list.append(tmp_db)
        logger.debug(f"Dump source {source} to temporary database: {tmp_db}")
        cmd = self.sqlite_cmd.copy()
        cmd.extend([source, f'.backup "{tmp_db}"'])
        subprocess.check_call(
            cmd,
            stdout=subprocess.DEVNULL,
            stderr=self.stream_err,
        )
        if output is None:
            tmp_file = open(tmp_db, mode="rb")
            self.tmp_file_list.append(tmp_file)
            return self._compress(
                in_stream=tmp_file,
                out_stream=subprocess.PIPE,
            )
        with open(tmp_db, mode="rb") as tmp_file:
            with open(output, mode="wb") as out_file:
                p = self._compress(
                    in_stream=tmp_file,
                    out_stream=out_file,
                )
                p.wait()
                return p

    def _cleanup(self):
        """Удаление временных объектов."""
        for tmp_file in self.tmp_file_list:
            tmp_file.close()
        self.tmp_file_list.clear()
        for tmp_db in self.tmp_db_list:
            if os.path.exists(tmp_db):
                logger.debug(f"Remove temporary database: {tmp_db}")
                os.remove(tmp_db)
        self.tmp_db_list.clear()
